@@ -22,4 +22,6 @@ void resume_all(void);
 void per_loop_maint(void);
 bool suspend_socket(PgSocket *sk, bool force)  _MUSTCHECK;
 void kill_pool(PgPool *pool);
+void kill_peer_pool(PgPool *pool);
 void kill_database(PgDatabase *db);
+void kill_peer(PgDatabase *db);
