@@ -18,6 +18,7 @@
 
 #include <usual/logging.h>
 #include <usual/string.h>
+#include <usual/cfparser.h>
 
 /*
  * logging about specific socket
@@ -29,18 +30,18 @@ int log_socket_prefix(enum LogLevel lev, void *ctx, char *dst, unsigned int dstl
 #define slog_info(sk, args...) log_generic(LG_INFO, sk, ## args)
 #define slog_debug(sk, args...) do { \
 		if (unlikely(cf_verbose > 0)) \
-			log_generic(LG_DEBUG, sk, ## args); \
-	} while (0)
+		log_generic(LG_DEBUG, sk, ## args); \
+} while (0)
 #define slog_noise(sk, args...) do { \
 		if (unlikely(cf_verbose > 1)) \
-			log_generic(LG_NOISE, sk, ## args); \
-	} while (0)
+		log_generic(LG_NOISE, sk, ## args); \
+} while (0)
 
 /*
  * password tools
  */
 #define MD5_PASSWD_LEN  35
-void pg_md5_encrypt(const char *part1, const char *part2, size_t p2len, char *dest);
+bool pg_md5_encrypt(const char *part1, const char *part2, size_t p2len, char *dest) _MUSTCHECK;
 void get_random_bytes(uint8_t *dest, int len);
 
 const char *bin2hex(const uint8_t *src, unsigned srclen, char *dst, unsigned dstlen);
@@ -58,7 +59,16 @@ void safe_evtimer_add(struct event *ev, struct timeval *tv);
 
 /* log truncated strings */
 #define safe_strcpy(dst, src, dstlen) do { \
-	size_t needed = strlcpy(dst, src, dstlen); \
-	if (unlikely(needed >= (dstlen))) \
+		size_t needed = strlcpy(dst, src, dstlen); \
+		if (unlikely(needed >= (dstlen))) \
 		log_warning("bug in %s:%d - string truncated", __FILE__, __LINE__); \
 } while (0)
+
+/* parser setter and getter for parsing auth_dbname parameters using "cfparser" */
+#define CF_AUTHDB       { cf_set_authdb, cf_get_str }
+bool cf_set_authdb(struct CfValue *cv, const char *value);
+
+/* reserved database name checking */
+bool check_reserved_database(const char *value);
+
+bool strings_equal(const char *str_left, const char *str_right) _MUSTCHECK;
