@@ -844,7 +844,7 @@ void kill_database(PgDatabase *db)
 	free(db->host);
 
 	if (db->forced_user_credentials)
-		slab_free(user_cache, db->forced_user_credentials);
+		slab_free(credentials_cache, db->forced_user_credentials);
 	free(db->connect_query);
 	if (db->inactive_time) {
 		statlist_remove(&autodatabase_idle_list, &db->head);
