@@ -21,6 +21,7 @@
  */
 
 #include "bouncer.h"
+#include "usual/time.h"
 
 #include <usual/fileutil.h>
 #include <usual/string.h>
@@ -121,35 +122,24 @@ static char * cstr_get_pair(char *p,
 }
 
 /*
- * Same as strcmp, but handles NULLs. If both sides are NULL, returns "true".
+ * Free the old value and set the new value
  */
-static bool strings_equal(const char *str_left, const char *str_right)
+static bool set_param_value(char **old_value, const char *new_value)
 {
-	if (str_left == NULL && str_right == NULL)
+	if (strings_equal(*old_value, new_value))
 		return true;
 
-	if (str_left == NULL || str_right == NULL)
-		return false;
+	if (*old_value)
+		free(*old_value);
 
-	return strcmp(str_left, str_right) == 0;
-}
-
-static bool set_auth_dbname(PgDatabase *db, const char *new_auth_dbname)
-{
-	if (strings_equal(db->auth_dbname, new_auth_dbname))
-		return true;
-
-	if (db->auth_dbname)
-		free(db->auth_dbname);
-
-	if (new_auth_dbname) {
-		db->auth_dbname = strdup(new_auth_dbname);
-		if (!db->auth_dbname) {
-			log_error("auth_dbname %s could not be set for database %s, out of memory", new_auth_dbname, db->name);
+	if (new_value) {
+		*old_value = strdup(new_value);
+		if (!(*old_value)) {
+			log_error("out of memory");
 			return false;
 		}
 	} else {
-		db->auth_dbname = NULL;
+		*old_value = NULL;
 	}
 
 	return true;
@@ -263,6 +253,7 @@ bool parse_database(void *base, const char *name, const char *connstr)
 	int min_pool_size = -1;
 	int res_pool_size = -1;
 	int max_db_connections = -1;
+	usec_t server_lifetime = 0;
 	int dbname_ofs;
 	int pool_mode = POOL_INHERIT;
 
@@ -279,6 +270,7 @@ bool parse_database(void *base, const char *name, const char *connstr)
 	char *timezone = NULL;
 	char *connect_query = NULL;
 	char *appname = NULL;
+	char *auth_query = NULL;
 
 	cv.value_p = &pool_mode;
 	cv.extra = (const void *)pool_mode_map;
@@ -344,6 +336,8 @@ bool parse_database(void *base, const char *name, const char *connstr)
 			res_pool_size = atoi(val);
 		} else if (strcmp("max_db_connections", key) == 0) {
 			max_db_connections = atoi(val);
+		} else if (strcmp("server_lifetime", key) == 0) {
+			server_lifetime = atoi(val) * USEC;
 		} else if (strcmp("pool_mode", key) == 0) {
 			if (!cf_set_lookup(&cv, val)) {
 				log_error("invalid pool mode: %s", val);
@@ -357,6 +351,8 @@ bool parse_database(void *base, const char *name, const char *connstr)
 			}
 		} else if (strcmp("application_name", key) == 0) {
 			appname = val;
+		} else if (strcmp("auth_query", key) == 0) {
+			auth_query = val;
 		} else {
 			log_error("unrecognized connection parameter: %s", key);
 			goto fail;
@@ -384,16 +380,18 @@ bool parse_database(void *base, const char *name, const char *connstr)
 			changed = true;
 		} else if (port != db->port) {
 			changed = true;
-		} else if (username && !db->forced_user) {
+		} else if (username && !db->forced_user_credentials) {
 			changed = true;
-		} else if (username && strcmp(username, db->forced_user->name) != 0) {
+		} else if (username && strcmp(username, db->forced_user_credentials->name) != 0) {
 			changed = true;
-		} else if (!username && db->forced_user) {
+		} else if (!username && db->forced_user_credentials) {
 			changed = true;
 		} else if (!strings_equal(connect_query, db->connect_query)) {
 			changed = true;
 		} else if (!strings_equal(db->auth_dbname, auth_dbname)) {
 			changed = true;
+		} else if (!strings_equal(db->auth_query, auth_query)) {
+			changed = true;
 		}
 		if (changed)
 			tag_database_dirty(db);
@@ -407,10 +405,14 @@ bool parse_database(void *base, const char *name, const char *connstr)
 	db->res_pool_size = res_pool_size;
 	db->pool_mode = pool_mode;
 	db->max_db_connections = max_db_connections;
+	db->server_lifetime = server_lifetime;
 	free(db->connect_query);
 	db->connect_query = connect_query;
 
-	if (!set_auth_dbname(db, auth_dbname))
+	if (!set_param_value(&db->auth_dbname, auth_dbname))
+		goto fail;
+
+	if (!set_param_value(&db->auth_query, auth_query))
 		goto fail;
 
 	if (db->startup_params) {
@@ -448,19 +450,19 @@ bool parse_database(void *base, const char *name, const char *connstr)
 	}
 
 	if (auth_username != NULL) {
-		db->auth_user = find_user(auth_username);
-		if (!db->auth_user) {
-			db->auth_user = add_user(auth_username, "");
+		db->auth_user_credentials = find_global_credentials(auth_username);
+		if (!db->auth_user_credentials) {
+			db->auth_user_credentials = add_global_credentials(auth_username, "");
 		}
-	} else if (db->auth_user) {
-		db->auth_user = NULL;
+	} else if (db->auth_user_credentials) {
+		db->auth_user_credentials = NULL;
 	}
 
 	/* if user is forced, create fake object for it */
 	if (username != NULL) {
-		if (!force_user(db, username, password))
+		if (!force_user_credentials(db, username, password))
 			log_warning("db setup failed, trying to continue");
-	} else if (db->forced_user) {
+	} else if (db->forced_user_credentials) {
 		log_warning("losing forced user not supported,"
 			    " keeping old setting");
 	}
@@ -478,9 +480,10 @@ fail:
 bool parse_user(void *base, const char *name, const char *connstr)
 {
 	char *p, *key, *val, *tmp_connstr;
-	PgUser *user;
+	PgGlobalUser *user;
 	struct CfValue cv;
 	int pool_mode = POOL_INHERIT;
+	int pool_size = -1;
 	int max_user_connections = -1;
 
 
@@ -508,6 +511,8 @@ bool parse_user(void *base, const char *name, const char *connstr)
 				log_error("invalid pool mode: %s", val);
 				goto fail;
 			}
+		} else if (strcmp("pool_size", key) == 0) {
+			pool_size = atoi(val);
 		} else if (strcmp("max_user_connections", key) == 0) {
 			max_user_connections = atoi(val);
 		} else {
@@ -516,9 +521,9 @@ bool parse_user(void *base, const char *name, const char *connstr)
 		}
 	}
 
-	user = find_user(name);
+	user = find_global_user(name);
 	if (!user) {
-		user = add_user(name, "");
+		user = add_global_user(name, "");
 		if (!user) {
 			log_error("cannot create user, no memory?");
 			goto fail;
@@ -526,6 +531,7 @@ bool parse_user(void *base, const char *name, const char *connstr)
 	}
 
 	user->pool_mode = pool_mode;
+	user->pool_size = pool_size;
 	user->max_user_connections = max_user_connections;
 
 	free(tmp_connstr);
@@ -566,18 +572,24 @@ static void copy_quoted(char *dst, const char *src, int len)
 	*dst = 0;
 }
 
-static void unquote_add_user(const char *username, const char *password)
+/* This function is only called when parsing the auth file, so
+   all users added by this function do not have a dynamic password,
+   by definition. If the password is empty, so be it. */
+static void unquote_add_authfile_user(const char *username, const char *password)
 {
 	char real_user[MAX_USERNAME];
 	char real_passwd[MAX_PASSWORD];
-	PgUser *user;
+	PgGlobalUser *user;
 
 	copy_quoted(real_user, username, sizeof(real_user));
 	copy_quoted(real_passwd, password, sizeof(real_passwd));
 
-	user = add_user(real_user, real_passwd);
-	if (!user)
+	user = add_global_user(real_user, real_passwd);
+	if (!user) {
 		log_warning("cannot create user, no memory");
+		return;
+	}
+	user->credentials.dynamic_passwd = false;
 }
 
 static bool auth_loaded(const char *fn)
@@ -619,12 +631,11 @@ bool loader_users_check(void)
 
 static void disable_users(void)
 {
-	PgUser *user;
 	struct List *item;
 
 	statlist_for_each(item, &user_list) {
-		user = container_of(item, PgUser, head);
-		user->passwd[0] = 0;
+		PgGlobalUser *user = container_of(item, PgGlobalUser, head);
+		user->credentials.passwd[0] = 0;
 	}
 }
 
@@ -695,7 +706,7 @@ bool load_auth_file(const char *fn)
 		*p++ = 0;	/* tag password end */
 
 		/* send them away */
-		unquote_add_user(user, password);
+		unquote_add_authfile_user(user, password);
 
 		/* skip rest of the line */
 		while (*p && *p != '\n') p++;
