@@ -102,8 +102,10 @@ bool get_header(struct MBuf *data, PktHdr *pkt)
 			type = PKT_SSLREQ;
 		} else if (code == PKT_GSSENCREQ) {
 			type = PKT_GSSENCREQ;
-		} else if ((code >> 16) == 3 && (code & 0xFFFF) < 2) {
-			type = PKT_STARTUP;
+		} else if (code >= PKT_STARTUP_V3 && code < PKT_STARTUP_V3_UNSUPPORTED) {
+			type = PKT_STARTUP_V3;
+		} else if (code >= PKT_STARTUP_V3_UNSUPPORTED && code < PKT_STARTUP_V4) {
+			type = PKT_STARTUP_V3_UNSUPPORTED;
 		} else if (code == PKT_STARTUP_V2) {
 			type = PKT_STARTUP_V2;
 		} else {
@@ -268,19 +270,18 @@ bool welcome_client(PgSocket *client)
 	 * If pgbouncer peering is enabled we change some of the random bits of the
 	 * cancel key to non random values, otherwise the peering feature cannot be
 	 * implemented in an efficient way. This reduces the randomness of the key
-	 * somewhat, but it still leaves us with 46 bits of randomness. This should
+	 * somewhat, but it still leaves us with 45 bits of randomness. This should
 	 * be enough for all practical attacks to be mitigated (there are still
-	 * ~70 trillion random combinations of these bits).
+	 * ~35 trillion random combinations of these bits).
 	 */
-	if (cf_peer_id > 0)
-	{
+	if (cf_peer_id > 0) {
 		/*
-		 * The first two bytes represent the peer id. Pgbouncers that are
+		 * The 2nd and 3rd byte represent the peer id. Pgbouncers that are
 		 * peered with this one can forward the request to us by reading this
 		 * peer id when they receive this cancellation.
 		 */
-		client->cancel_key[0] = cf_peer_id & 0xFF;
-		client->cancel_key[1] = (cf_peer_id >> 8) & 0xFF;
+		client->cancel_key[1] = cf_peer_id & 0xFF;
+		client->cancel_key[2] = (cf_peer_id >> 8) & 0xFF;
 
 		/*
 		 * Initially we set the two TTL mask bits to a 1, so that the cancel
@@ -289,6 +290,17 @@ bool welcome_client(PgSocket *client)
 		client->cancel_key[7] |= CANCELLATION_TTL_MASK;
 	}
 
+	/*
+	 * The first 32 bits of the cancel_key are considered a PID. Since
+	 * actual PIDs are always positive we clear the sign bit. Most clients
+	 * work fine when receiving a negative number in this PID part, but it
+	 * turned out that pg_basebackup did not. This is fixed in
+	 * pg_basebackup, but to avoid similar future problems with other
+	 * clients we clear the sign bit. See this thread for more details:
+	 * https://www.postgresql.org/message-id/flat/CAGECzQQOGvYfp8ziF4fWQ_o8s2K7ppaoWBQnTmdakn3s-4Z%3D5g%40mail.gmail.com
+	 */
+	client->cancel_key[0] &= 0x7F;
+
 	pktbuf_write_BackendKeyData(msg, client->cancel_key);
 
 	/* finish */
@@ -311,18 +323,18 @@ bool welcome_client(PgSocket *client)
  * Password authentication for server
  */
 
-static PgUser *get_srv_psw(PgSocket *server)
+static PgCredentials *get_srv_psw(PgSocket *server)
 {
 	PgDatabase *db = server->pool->db;
-	PgUser *user = server->pool->user;
+	PgCredentials *credentials = server->pool->user_credentials;
 
 	/* if forced user without password, use userlist psw */
-	if (!user->passwd[0] && db->forced_user) {
-		PgUser *u2 = find_user(user->name);
-		if (u2)
-			return u2;
+	if (!credentials->passwd[0] && db->forced_user_credentials) {
+		PgCredentials *c2 = find_global_credentials(credentials->name);
+		if (c2)
+			return c2;
 	}
-	return user;
+	return credentials;
 }
 
 /* actual packet send */
@@ -335,25 +347,26 @@ static bool send_password(PgSocket *server, const char *enc_psw)
 
 static bool login_clear_psw(PgSocket *server)
 {
-	PgUser *user = get_srv_psw(server);
+	PgCredentials *credentials = get_srv_psw(server);
 	slog_debug(server, "P: send clear password");
-	return send_password(server, user->passwd);
+	return send_password(server, credentials->passwd);
 }
 
 static bool login_md5_psw(PgSocket *server, const uint8_t *salt)
 {
 	char txt[MD5_PASSWD_LEN + 1], *src;
-	PgUser *user = get_srv_psw(server);
+	PgCredentials *credentials = get_srv_psw(server);
 
 	slog_debug(server, "P: send md5 password");
 
-	switch (get_password_type(user->passwd)) {
+	switch (get_password_type(credentials->passwd)) {
 	case PASSWORD_TYPE_PLAINTEXT:
-		pg_md5_encrypt(user->passwd, user->name, strlen(user->name), txt);
+		if (!pg_md5_encrypt(credentials->passwd, credentials->name, strlen(credentials->name), txt))
+			return false;
 		src = txt + 3;
 		break;
 	case PASSWORD_TYPE_MD5:
-		src = user->passwd + 3;
+		src = credentials->passwd + 3;
 		break;
 	default:
 		slog_error(server, "cannot do MD5 authentication: wrong password type");
@@ -361,23 +374,24 @@ static bool login_md5_psw(PgSocket *server, const uint8_t *salt)
 		return false;
 	}
 
-	pg_md5_encrypt(src, (char *)salt, 4, txt);
+	if (!pg_md5_encrypt(src, (char *)salt, 4, txt))
+		return false;
 
 	return send_password(server, txt);
 }
 
 static bool login_scram_sha_256(PgSocket *server)
 {
-	PgUser *user = get_srv_psw(server);
+	PgCredentials *credentials = get_srv_psw(server);
 	bool res;
 	char *client_first_message = NULL;
 
-	switch (get_password_type(user->passwd)) {
+	switch (get_password_type(credentials->passwd)) {
 	case PASSWORD_TYPE_PLAINTEXT:
 		/* ok */
 		break;
 	case PASSWORD_TYPE_SCRAM_SHA_256:
-		if (!user->has_scram_keys) {
+		if (!credentials->has_scram_keys) {
 			slog_error(server, "cannot do SCRAM authentication: password is SCRAM secret but client authentication did not provide SCRAM keys");
 			kill_pool_logins(server->pool, NULL, "server login failed: wrong password type");
 			return false;
@@ -389,8 +403,7 @@ static bool login_scram_sha_256(PgSocket *server)
 		return false;
 	}
 
-	if (server->scram_state.client_nonce)
-	{
+	if (server->scram_state.client_nonce) {
 		slog_error(server, "protocol error: duplicate AuthenticationSASL message from server");
 		return false;
 	}
@@ -409,7 +422,7 @@ static bool login_scram_sha_256(PgSocket *server)
 
 static bool login_scram_sha_256_cont(PgSocket *server, unsigned datalen, const uint8_t *data)
 {
-	PgUser *user = get_srv_psw(server);
+	PgCredentials *credentials = get_srv_psw(server);
 	char *ibuf = NULL;
 	char *input;
 	char *server_nonce;
@@ -419,14 +432,12 @@ static bool login_scram_sha_256_cont(PgSocket *server, unsigned datalen, const u
 	bool res;
 	char *client_final_message = NULL;
 
-	if (!server->scram_state.client_nonce)
-	{
+	if (!server->scram_state.client_nonce) {
 		slog_error(server, "protocol error: AuthenticationSASLContinue without prior AuthenticationSASL");
 		return false;
 	}
 
-	if (server->scram_state.server_first_message)
-	{
+	if (server->scram_state.server_first_message) {
 		slog_error(server, "SCRAM exchange protocol error: received second AuthenticationSASLContinue");
 		return false;
 	}
@@ -444,7 +455,7 @@ static bool login_scram_sha_256_cont(PgSocket *server, unsigned datalen, const u
 		goto failed;
 
 	client_final_message = build_client_final_message(&server->scram_state,
-							  user, server_nonce,
+							  credentials, server_nonce,
 							  salt, saltlen, iterations);
 
 	free(salt);
@@ -465,13 +476,12 @@ failed:
 
 static bool login_scram_sha_256_final(PgSocket *server, unsigned datalen, const uint8_t *data)
 {
-	PgUser *user = get_srv_psw(server);
+	PgCredentials *credentials = get_srv_psw(server);
 	char *ibuf = NULL;
 	char *input;
 	char ServerSignature[SHA256_DIGEST_LENGTH];
 
-	if (!server->scram_state.server_first_message)
-	{
+	if (!server->scram_state.server_first_message) {
 		slog_error(server, "protocol error: AuthenticationSASLFinal without prior AuthenticationSASLContinue");
 		return false;
 	}
@@ -487,11 +497,10 @@ static bool login_scram_sha_256_final(PgSocket *server, unsigned datalen, const
 	if (!read_server_final_message(server, input, ServerSignature))
 		goto failed;
 
-	if (!verify_server_signature(&server->scram_state, user, ServerSignature))
-	{
+	if (!verify_server_signature(&server->scram_state, credentials, ServerSignature)) {
 		slog_error(server, "invalid server signature");
 		kill_pool_logins(server->pool, NULL, "server login failed: invalid server signature");
-		return false;
+		goto failed;
 	}
 
 	free(ibuf);
@@ -550,8 +559,9 @@ bool answer_authreq(PgSocket *server, PktHdr *pkt)
 		if (!selected_mechanism) {
 			slog_error(server, "none of the server's SASL authentication mechanisms are supported");
 			kill_pool_logins(server->pool, NULL, "server login failed: none of the server's SASL authentication mechanisms are supported");
-		} else
+		} else {
 			res = login_scram_sha_256(server);
+		}
 		break;
 	}
 	case AUTH_SASL_CONT:
@@ -589,15 +599,69 @@ bool answer_authreq(PgSocket *server, PktHdr *pkt)
 
 bool send_startup_packet(PgSocket *server)
 {
-	PgDatabase *db = server->pool->db;
-	const char *username = server->pool->user->name;
-	PktBuf *pkt;
+	PgPool *pool = server->pool;
+	PgDatabase *db = pool->db;
+	const char *username = server->pool->user_credentials->name;
+	PktBuf *pkt = pktbuf_temp();
+	PgSocket *client = NULL;
 
-	pkt = pktbuf_temp();
-	pktbuf_write_StartupMessage(pkt, username,
-				    db->startup_params->buf,
-				    db->startup_params->write_pos);
-	return pktbuf_send_immediate(pkt, server);
+	pktbuf_start_packet(pkt, PKT_STARTUP_V3);
+	pktbuf_put_bytes(pkt, db->startup_params->buf, db->startup_params->write_pos);
+
+	/*
+	 * If the next client in the list is a replication connection, we need
+	 * to do some special stuff for it.
+	 */
+	client = first_socket(&pool->waiting_client_list);
+	if (client && client->replication) {
+		server->replication = client->replication;
+		pktbuf_put_string(pkt, "replication");
+		slog_debug(server, "send_startup_packet: creating replication connection");
+		pktbuf_put_string(pkt, replication_type_parameters[server->replication]);
+
+		/*
+		 * For a replication connection we apply the varcache in the
+		 * startup instead of through SET commands after connecting.
+		 * The main reason to do so is because physical replication
+		 * connections don't allow SET commands. A second reason is
+		 * because it allows us to skip running the SET logic
+		 * completely, which normally requires waiting on multiple
+		 * server responses. This SET logic is normally executed in the
+		 * codepath where we link the client to the server
+		 * (find_server), but because we link the client here already
+		 * we don't run that code for replication connections. Adding
+		 * the varcache parameters to the startup message allows us to
+		 * skip the dance that involves sending Query packets and
+		 * waiting for responses.
+		 */
+		varcache_apply_startup(pkt, client);
+		if (client->startup_options) {
+			pktbuf_put_string(pkt, "options");
+			pktbuf_put_string(pkt, client->startup_options);
+		}
+	}
+
+	pktbuf_put_string(pkt, "user");
+	pktbuf_put_string(pkt, username);
+	pktbuf_put_string(pkt, "");	/* terminator required in StartupMessage */
+	pktbuf_finish_packet(pkt);
+
+	if (!pktbuf_send_immediate(pkt, server)) {
+		return false;
+	}
+
+	if (server->replication) {
+		/*
+		 * We link replication connections to a client directly when they are
+		 * created. One reason for is because the startup parameters need to be
+		 * forwarded, because physical replication connections don't allow SET
+		 * commands. Another reason is so that we don't need a separate state.
+		 */
+		client->link = server;
+		server->link = client;
+	}
+
+	return true;
 }
 
 bool send_sslreq_packet(PgSocket *server)
@@ -633,15 +697,13 @@ int scan_text_result(struct MBuf *pkt, const char *tupdesc, ...)
 
 		if (i < ncol) {
 			if (!mbuf_get_uint32be(pkt, &len)) {
-				va_end(ap);
-				return -1;
+				goto failed;
 			}
 			if ((int32_t)len < 0) {
 				val = NULL;
 			} else {
 				if (!mbuf_get_chars(pkt, len, &val)) {
-					va_end(ap);
-					return -1;
+					goto failed;
 				}
 			}
 
@@ -688,14 +750,15 @@ int scan_text_result(struct MBuf *pkt, const char *tupdesc, ...)
 				int newlen;
 				if (strncmp(val, "\\x", 2) != 0) {
 					log_warning("invalid bytea value");
-					return -1;
+					goto failed;
 				}
 
 				newlen = (len - 2) / 2;
 				*len_p = newlen;
 				*bytes_p = malloc(newlen);
-				if (!(*bytes_p))
-					return -1;
+				if (!(*bytes_p)) {
+					goto failed;
+				}
 				for (int j = 0; j < newlen; j++) {
 					unsigned int b;
 					sscanf(val + 2 + 2 * j, "%2x", &b);
@@ -714,4 +777,7 @@ int scan_text_result(struct MBuf *pkt, const char *tupdesc, ...)
 	va_end(ap);
 
 	return ncol;
+failed:
+	va_end(ap);
+	return -1;
 }
