@@ -12,6 +12,7 @@ import asyncio
 import os
 import platform
 import re
+import shlex
 import signal
 import socket
 import sys
@@ -21,6 +22,8 @@ from tempfile import gettempdir
 
 import filelock
 import psycopg
+import psycopg.sql
+from psycopg import sql
 
 TEST_DIR = Path(os.path.dirname(os.path.realpath(__file__)))
 os.chdir(TEST_DIR)
@@ -83,28 +86,45 @@ def eprint(*args, **kwargs):
     print(*args, file=sys.stderr, **kwargs)
 
 
-def run(command, *args, check=True, shell=True, silent=False, **kwargs):
+def run(command, *args, check=True, shell=None, silent=False, **kwargs):
     """run runs the given command and prints it to stderr"""
 
+    if shell is None:
+        shell = isinstance(command, str)
+
+    if not shell:
+        command = list(map(str, command))
+
     if not silent:
-        eprint(f"+ {command} ")
+        if shell:
+            eprint(f"+ {command}")
+        else:
+            # We could normally use shlex.join here, but it's not available in
+            # Python 3.6 which we still like to support
+            unsafe_string_cmd = " ".join(map(shlex.quote, command))
+            eprint(f"+ {unsafe_string_cmd}")
     if silent:
         kwargs.setdefault("stdout", subprocess.DEVNULL)
     return subprocess.run(command, *args, check=check, shell=shell, **kwargs)
 
 
-def sudo(command, *args, shell=True, **kwargs):
+def sudo(command, *args, shell=None, **kwargs):
     """
     A version of run that prefixes the command with sudo when the process is
     not already run as root
     """
     effective_user_id = os.geteuid()
+
     if effective_user_id == 0:
         return run(command, *args, shell=shell, **kwargs)
+
+    if shell is None:
+        shell = isinstance(command, str)
+
     if shell:
         return run(f"sudo {command}", *args, shell=shell, **kwargs)
     else:
-        return run(["sudo", *command])
+        return run(["sudo", *command], *args, shell=shell, **kwargs)
 
 
 def capture(command, *args, stdout=subprocess.PIPE, encoding="utf-8", **kwargs):
@@ -132,6 +152,7 @@ def get_max_password_length():
     return max_password_length
 
 
+PKT_BUF_SIZE = 4096
 MAX_PASSWORD_LENGTH = get_max_password_length()
 LONG_PASSWORD = "a" * MAX_PASSWORD_LENGTH
 
@@ -162,6 +183,30 @@ PORT_UPPER_BOUND = 32768
 next_port = PORT_LOWER_BOUND
 
 
+def cleanup_test_leftovers(*nodes):
+    """
+    Cleaning up test leftovers needs to be done in a specific order, because
+    some of these leftovers depend on others having been removed. They might
+    even depend on leftovers on other nodes being removed. So this takes a list
+    of nodes, so that we can clean up all test leftovers globally in the
+    correct order.
+    """
+    for node in nodes:
+        node.cleanup_subscriptions()
+
+    for node in nodes:
+        node.cleanup_publications()
+
+    for node in nodes:
+        node.cleanup_replication_slots()
+
+    for node in nodes:
+        node.cleanup_schemas()
+
+    for node in nodes:
+        node.cleanup_users()
+
+
 class PortLock:
     def __init__(self):
         global next_port
@@ -205,9 +250,19 @@ class QueryRunner:
         self.default_db = "postgres"
         self.default_user = "postgres"
 
+        # Used to track objects that we want to clean up at the end of a test
+        self.subscriptions = set()
+        self.publications = set()
+        self.replication_slots = set()
+        self.schemas = set()
+        self.users = set()
+
     def set_default_connection_options(self, options):
+        """Sets the default connection options on the given options dictionary"""
         options.setdefault("dbname", self.default_db)
         options.setdefault("user", self.default_user)
+        options.setdefault("host", self.host)
+        options.setdefault("port", self.port)
         if ENABLE_VALGRIND:
             # If valgrind is enabled PgBouncer is a significantly slower to
             # respond to connection requests, so we wait a little longer.
@@ -217,16 +272,19 @@ class QueryRunner:
         # Always required for Ubuntu 18.04, but also needed for any tests
         # involving the varcache_change database. The difference between the
         # client_encoding specified in the config and client_encoding by the
-        # client will force a varcache change when a connectino is given.
+        # client will force a varcache change when a connection is given.
         options.setdefault("client_encoding", "UTF8")
+        return options
+
+    def make_conninfo(self, **kwargs) -> str:
+        self.set_default_connection_options(kwargs)
+        return psycopg.conninfo.make_conninfo(**kwargs)
 
     def conn(self, *, autocommit=True, **kwargs):
         """Open a psycopg connection to this server"""
         self.set_default_connection_options(kwargs)
         conn = psycopg.connect(
             autocommit=autocommit,
-            host=self.host,
-            port=self.port,
             **kwargs,
         )
         conn.add_notice_handler(notice_handler)
@@ -237,8 +295,6 @@ class QueryRunner:
         self.set_default_connection_options(kwargs)
         return psycopg.AsyncConnection.connect(
             autocommit=autocommit,
-            host=self.host,
-            port=self.port,
             **kwargs,
         )
 
@@ -274,6 +330,12 @@ class QueryRunner:
         """
         with self.cur(**kwargs) as cur:
             cur.execute(query, params=params)
+            try:
+                return cur.fetchall()
+            except psycopg.ProgrammingError as e:
+                if "the last operation didn't produce a result" == str(e):
+                    return None
+                raise
 
     def sql_value(self, query, params=None, **kwargs):
         """Run an SQL query that returns a single cell and return this value
@@ -481,6 +543,116 @@ class QueryRunner:
             sudo(f"tc filter del dev lo parent 1: prio {self.port}")
             pass
 
+    def create_user(self, name, args: typing.Optional[psycopg.sql.Composable] = None):
+        self.users.add(name)
+        if args is None:
+            args = sql.SQL("")
+        self.sql(sql.SQL("CREATE USER {} {}").format(sql.Identifier(name), args))
+
+    def create_schema(self, name, dbname=None):
+        dbname = dbname or self.default_db
+        self.schemas.add((dbname, name))
+        self.sql(sql.SQL("CREATE SCHEMA {}").format(sql.Identifier(name)))
+
+    def create_publication(self, name: str, args: psycopg.sql.Composable, dbname=None):
+        dbname = dbname or self.default_db
+        self.publications.add((dbname, name))
+        self.sql(sql.SQL("CREATE PUBLICATION {} {}").format(sql.Identifier(name), args))
+
+    def create_logical_replication_slot(self, name, plugin):
+        self.replication_slots.add(name)
+        self.sql(
+            "SELECT pg_catalog.pg_create_logical_replication_slot(%s,%s)",
+            (name, plugin),
+        )
+
+    def create_physical_replication_slot(self, name):
+        self.replication_slots.add(name)
+        self.sql(
+            "SELECT pg_catalog.pg_create_physical_replication_slot(%s)",
+            (name,),
+        )
+
+    def create_subscription(self, name: str, args: psycopg.sql.Composable, dbname=None):
+        dbname = dbname or self.default_db
+        self.subscriptions.add((dbname, name))
+        self.sql(
+            sql.SQL("CREATE SUBSCRIPTION {} {}").format(sql.Identifier(name), args)
+        )
+
+    def cleanup_users(self):
+        for user in self.users:
+            self.sql(sql.SQL("DROP USER IF EXISTS {}").format(sql.Identifier(user)))
+
+    def cleanup_schemas(self):
+        for dbname, schema in self.schemas:
+            self.sql(
+                sql.SQL("DROP SCHEMA IF EXISTS {} CASCADE").format(
+                    sql.Identifier(schema)
+                ),
+                dbname=dbname,
+            )
+
+    def cleanup_publications(self):
+        for dbname, publication in self.publications:
+            self.sql(
+                sql.SQL("DROP PUBLICATION IF EXISTS {}").format(
+                    sql.Identifier(publication)
+                ),
+                dbname=dbname,
+            )
+
+    def cleanup_replication_slots(self):
+        for slot in self.replication_slots:
+            start = time.time()
+            while True:
+                try:
+                    self.sql(
+                        "SELECT pg_drop_replication_slot(slot_name) FROM pg_replication_slots WHERE slot_name = %s",
+                        (slot,),
+                    )
+                except psycopg.errors.ObjectInUse:
+                    if time.time() < start + 10:
+                        time.sleep(0.5)
+                        continue
+                    raise
+                break
+
+    def cleanup_subscriptions(self):
+        for dbname, subscription in self.subscriptions:
+            try:
+                self.sql(
+                    sql.SQL("ALTER SUBSCRIPTION {} DISABLE").format(
+                        sql.Identifier(subscription)
+                    ),
+                    dbname=dbname,
+                )
+            except psycopg.errors.UndefinedObject:
+                # Subscription didn't exist already
+                continue
+            self.sql(
+                sql.SQL("ALTER SUBSCRIPTION {} SET (slot_name = NONE)").format(
+                    sql.Identifier(subscription)
+                ),
+                dbname=dbname,
+            )
+            self.sql(
+                sql.SQL("DROP SUBSCRIPTION {}").format(sql.Identifier(subscription)),
+                dbname=dbname,
+            )
+
+    def debug(self):
+        print("Connect manually to:\n   ", repr(self.make_conninfo()))
+        print("Press Enter to continue running the test...")
+        input()
+
+    def psql_debug(self, **kwargs):
+        conninfo = self.make_conninfo(**kwargs)
+        run(
+            ["psql", conninfo],
+            silent=True,
+        )
+
 
 class Postgres(QueryRunner):
     def __init__(self, pgdata):
@@ -504,6 +676,26 @@ class Postgres(QueryRunner):
             pgconf.write("log_connections = on\n")
             pgconf.write("log_disconnections = on\n")
             pgconf.write("logging_collector = off\n")
+
+            # Allow CREATE SUBSCRIPTION to work
+            pgconf.write("wal_level = 'logical'\n")
+            # Faster logical replication status update so tests with logical replication
+            # run faster
+            pgconf.write("wal_receiver_status_interval = 1\n")
+
+            # Faster logical replication apply worker launch so tests with logical
+            # replication run faster. This is used in ApplyLauncherMain in
+            # src/backend/replication/logical/launcher.c.
+            pgconf.write("wal_retrieve_retry_interval = '250ms'\n")
+
+            # Make sure there's enough logical replication resources for our
+            # tests
+            if PG_MAJOR_VERSION >= 10:
+                pgconf.write("max_logical_replication_workers = 5\n")
+            pgconf.write("max_wal_senders = 5\n")
+            pgconf.write("max_replication_slots = 10\n")
+            pgconf.write("max_worker_processes = 20\n")
+
             # We need to make the log go to stderr so that the tests can
             # check what is being logged.  This should be the default, but
             # some packagings change the default configuration.
@@ -516,6 +708,9 @@ class Postgres(QueryRunner):
             # are the same.
             pgconf.write("extra_float_digits = 1\n")
 
+            # Make sure this is consistent across platforms
+            pgconf.write("datestyle = 'iso, mdy'\n")
+
     def pgctl(self, command, **kwargs):
         run(f"pg_ctl -w --pgdata {self.pgdata} {command}", **kwargs)
 
@@ -557,24 +752,24 @@ class Postgres(QueryRunner):
         process = await self.apgctl("-m fast restart")
         await process.communicate()
 
-    def nossl_access(self, dbname, auth_type):
+    def nossl_access(self, dbname, auth_type, user="all"):
         """Prepends a local non-SSL access to the HBA file"""
         with self.hba_path.open() as pghba:
             old_contents = pghba.read()
         with self.hba_path.open(mode="w") as pghba:
             if USE_UNIX_SOCKETS:
-                pghba.write(f"local {dbname}   all                {auth_type}\n")
-            pghba.write(f"hostnossl  {dbname}   all  127.0.0.1/32  {auth_type}\n")
-            pghba.write(f"hostnossl  {dbname}   all  ::1/128       {auth_type}\n")
+                pghba.write(f"local {dbname}   {user}                {auth_type}\n")
+            pghba.write(f"hostnossl  {dbname}   {user}  127.0.0.1/32  {auth_type}\n")
+            pghba.write(f"hostnossl  {dbname}   {user}  ::1/128       {auth_type}\n")
             pghba.write(old_contents)
 
-    def ssl_access(self, dbname, auth_type):
+    def ssl_access(self, dbname, auth_type, user="all"):
         """Prepends a local SSL access rule to the HBA file"""
         with self.hba_path.open() as pghba:
             old_contents = pghba.read()
         with self.hba_path.open(mode="w") as pghba:
-            pghba.write(f"hostssl  {dbname}   all  127.0.0.1/32  {auth_type}\n")
-            pghba.write(f"hostssl  {dbname}   all  ::1/128       {auth_type}\n")
+            pghba.write(f"hostssl  {dbname}   {user}  127.0.0.1/32  {auth_type}\n")
+            pghba.write(f"hostssl  {dbname}   {user}  ::1/128       {auth_type}\n")
             pghba.write(old_contents)
 
     @property
@@ -586,7 +781,7 @@ class Postgres(QueryRunner):
         return self.pgdata / "postgresql.conf"
 
     def commit_hba(self):
-        """Mark the current HBA contents as non-resetable by reset_hba"""
+        """Mark the current HBA contents as non-resettable by reset_hba"""
         with self.hba_path.open() as pghba:
             old_contents = pghba.read()
         with self.hba_path.open(mode="w") as pghba:
@@ -790,18 +985,38 @@ class Bouncer(QueryRunner):
         task"""
         return self.admin_runner.asql(query, **kwargs)
 
-    async def stop(self):
+    def running(self):
+        if self.process:
+            return self.process.poll() is None
+        if self.aprocess:
+            return self.aprocess.returncode is None
+        return False
+
+    async def wait_for_exit(self):
         if self.process is not None:
-            self.process.terminate()
             self.process.communicate()
             self.process.wait()
         if self.aprocess is not None:
-            self.aprocess.terminate()
             await self.aprocess.communicate()
             await self.aprocess.wait()
         self.process = None
         self.aprocess = None
 
+    async def stop(self):
+        if not WINDOWS:
+            self.sigquit()
+        else:
+            # Windows does not have SIGQUIT, so call terminate() twice to
+            # trigger fast exit
+            if self.process is not None:
+                self.process.terminate()
+                self.process.terminate()
+            if self.aprocess is not None:
+                self.aprocess.terminate()
+                self.aprocess.terminate()
+
+        await self.wait_for_exit()
+
     async def reboot(self):
         """Starts a new PgBouncer with the --reboot flag
 
@@ -835,13 +1050,28 @@ class Bouncer(QueryRunner):
             await self.wait_until_running()
             assert self.process.pid != old_pid
 
-    def sighup(self):
+    def send_signal(self, sig):
         if self.aprocess:
-            self.aprocess.send_signal(signal.SIGHUP)
+            self.aprocess.send_signal(sig)
         if self.process:
-            self.process.send_signal(signal.SIGHUP)
+            self.process.send_signal(sig)
+
+    def sighup(self):
+        self.send_signal(signal.SIGHUP)
         time.sleep(1)
 
+    def sigterm(self):
+        self.send_signal(signal.SIGTERM)
+
+    def sigint(self):
+        self.send_signal(signal.SIGINT)
+
+    def sigquit(self):
+        self.send_signal(signal.SIGQUIT)
+
+    def sigusr2(self):
+        self.send_signal(signal.SIGUSR2)
+
     def print_logs(self):
         print(f"\n\nBOUNCER_LOG {self.config_dir}\n")
 
@@ -873,8 +1103,11 @@ class Bouncer(QueryRunner):
             assert not failed_valgrind
 
     async def cleanup(self):
-        await self.stop()
-        self.print_logs()
+        try:
+            cleanup_test_leftovers(self)
+            await self.stop()
+        finally:
+            self.print_logs()
 
         if self.port_lock:
             self.port_lock.release()
